"""Plugin functionality for adding function transforms.

This package contains the abstract base classes for function transform plugins,
and the default function transforms that are part of `ropt`.
"""
